//---------------------------------------------------------------------------

#pragma hdrstop

#include "PluginInterface.h"
#include "Plugin.h"
#include "Log.h"
#include "Utilities.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

std::vector<DllInfo> Plugin::dlls;
std::set<class Plugin*> Plugin::instances;


void Plugin::EnumerateDlls(AnsiString dir)
{
	LOG(E_LOG_TRACE, "Enumerating device dlls...");
	Plugin::dlls.clear();
	WIN32_FIND_DATA file;
	if (dir[dir.Length()] != '\\')
		dir += "\\";
	AnsiString asSrchPath = dir + "*.dll";
	HANDLE hFind = FindFirstFile(asSrchPath.c_str(), &file);
	int hasfiles = (hFind != INVALID_HANDLE_VALUE);
	struct DllInfo dllinfo;

	while (hasfiles)
	{
		AnsiString filename = dir + file.cFileName;
		if(VerifyDll(filename, &dllinfo) == E_OK)
		{
			AddDll(dllinfo);
		} else
			LOG(E_LOG_TRACE, "Library %s not loaded", filename.c_str());
		hasfiles = FindNextFile(hFind, &file);
	}
	FindClose(hFind);
}

enum Plugin::E_LIB_STATUS Plugin::VerifyDll(AnsiString filename, struct DllInfo* const dllinfo)
{
#if 0
	// Suppressing "Cannot Find ThisLibrary.dll" or "Cannot Find SomeOtherRequired.dll" Error Box
	UINT uOldErrorMode = SetErrorMode(SEM_NOOPENFILEERRORBOX | SEM_FAILCRITICALERRORS);
	HINSTANCE hInstance = LoadLibrary(filename.c_str());
	SetErrorMode(uOldErrorMode);

	if ((int)hInstance == 2)
	{
		//MessageBox(NULL, "MYDLL.DLL does not exist", "ERROR", MB_OK);
		return false;
	}
	else if ((int)hInstance <= HINSTANCE_ERROR)
	{
		//MessageBox(NULL, "Other problems loading MYDLL.DLL, "ERROR", MB_OK);
		return false;
	}
#else
	// Windows error message when loading may actually be usefull when resolving additional dll dependencies
	HINSTANCE hInstance = LoadLibrary(filename.c_str());
	if (!hInstance)
	{
		int error = GetLastError();
		LOG(E_LOG_TRACE, "Failed to load %s library (error %d)", filename.c_str(), error);
		return E_LOADFAILED;
	}
#endif

	pfGetInterfaceDescription gid =
		(pfGetInterfaceDescription)GetProcAddress(hInstance, "GetInterfaceDescription");
	if (!gid)
	{
		LOG(E_LOG_TRACE, "Library %s DOES NOT look like a compatible plugin", filename.c_str());
		FreeLibrary(hInstance);
		return E_NOTVALID;
	}
	struct S_DLL_INTERFACE interf_version;
	gid(&interf_version);
	if (interf_version.majorVersion != DLL_INTERFACE_MAJOR_VERSION ||
		interf_version.minorVersion != DLL_INTERFACE_MINOR_VERSION)
	{
		LOG(E_LOG_TRACE, "Library %s has interface v.%d.%d instead of required %d.%d",
			filename.c_str(),
			interf_version.majorVersion, interf_version.minorVersion,
			DLL_INTERFACE_MAJOR_VERSION);
		FreeLibrary(hInstance);
		return E_INCOMPATIBLE_VER;
	}

	FreeLibrary(hInstance);
	dllinfo->name = filename;
	dllinfo->file_version = GetFileVer(dllinfo->name);
	dllinfo->ver.majorVersion = interf_version.majorVersion;
	dllinfo->ver.minorVersion = interf_version.minorVersion;
	return E_OK;
}

void Plugin::AddDll(const struct DllInfo &dllinfo)
{
	LOG(E_LOG_TRACE, "Adding %s", dllinfo.name.c_str());
	Plugin::dlls.push_back(dllinfo);
}

void __stdcall Plugin::OnSend(void *cookie, void *buffer, unsigned int buffersize)
{
	class Plugin *dev;
	dev = reinterpret_cast<class Plugin*>(cookie);
	if (instances.find(dev) == instances.end())
	{
		//LOG(E_LOG_TRACE, "OnData called with unknown cookie %p. No matching object.", dev);
		return;
	}
	if (dev->callbackSend)
		dev->callbackSend((short*)buffer, buffersize);
}

Plugin::Plugin(AnsiString asDllName):
	hInstance(NULL),
	filename(asDllName),
	callbackSend(NULL),
	dllGetInterfaceDescription(NULL),
	dllShowSettings(NULL),
	dllChangeConnectionState(NULL),
	dllOnRx(NULL),
	dllOnTx(NULL)
{
	LOG(E_LOG_TRACE, "Creating object using %s", asDllName.c_str());
	instances.insert(this);
}

void Plugin::Unload(void)
{
	std::set<class Plugin*>::iterator iter;
	for (iter=Plugin::instances.begin(); iter!=Plugin::instances.end(); ++iter)
	{
		Plugin* plugin = *iter;
		delete plugin;
	}
	Plugin::instances.clear();
}


Plugin::~Plugin()
{
	if (hInstance)
	{
		FreeLibrary(hInstance);
		hInstance = NULL;
	}
}

int Plugin::Load(void)
{
	hInstance = LoadLibrary(filename.c_str());
	if (!hInstance)
		return 1;
	dllSetCallbacks = (pfSetCallbacks)GetProcAddress(hInstance, "SetCallbacks");
	dllShowSettings = (pfShowSettings)GetProcAddress(hInstance, "ShowSettings");
	dllChangeConnectionState = (pfChangeConnectionState)GetProcAddress(hInstance, "ChangeConnectionState");
	dllOnRx = (pfOnRx)GetProcAddress(hInstance, "OnRx");
	dllOnTx = (pfOnTx)GetProcAddress(hInstance, "OnTx");

	if ((dllSetCallbacks && dllShowSettings) == 0)
	{
		LOG(E_LOG_TRACE, "Dll load failed. Some of the required functions are missing");
		return 2;
	}

	dllSetCallbacks(this, &OnSend);
	return 0;
}

